@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2021 by The Linux Foundation
+# Copyright (C) 2021-2022 by The Linux Foundation
 # SPDX-License-Identifier: MIT-0
 #
 __author__ = 'Konstantin Ryabitsev <konstantin@linuxfoundation.org>'
@@ -28,7 +28,8 @@ from io import BytesIO
 logger = logging.getLogger(__name__)
 
 # Overridable via [patatt] parameters
-GPGBIN = 'gpg'
+GPGBIN = None
+SSHKBIN = None
 
 # Hardcoded defaults
 DEVSIG_HDR = b'X-Developer-Signature'
@@ -36,6 +37,7 @@ DEVKEY_HDR = b'X-Developer-Key'
 
 # Result and severity levels
 RES_VALID = 0
+RES_NOSIG = 4
 RES_NOKEY = 8
 RES_ERROR = 16
 RES_BADSIG = 32
@@ -45,9 +47,11 @@ OPT_HDRS = [b'message-id']
 
 # Quick cache for key info
 KEYCACHE = dict()
+# Quick cache for config settings
+CONFIGCACHE = dict()
 
 # My version
-__VERSION__ = '0.4.9'
+__VERSION__ = '0.6.2'
 MAX_SUPPORTED_FORMAT_VERSION = 1
 
 
@@ -204,6 +208,13 @@ class DevsigHeader:
         pts = self.hval.rsplit(b'b=', 1)
         dshdr = pts[0] + b'b='
         bdata = re.sub(rb'\s*', b'', pts[1])
+        # Calculate our own digest
+        hashed = hashlib.sha256()
+        # Add in our _headervals first (they aready have CRLF endings)
+        hashed.update(b''.join(self._headervals))
+        # and the devsig header now, without the trailing CRLF
+        hashed.update(DEVSIG_HDR.lower() + b':' + dshdr)
+        vdigest = hashed.digest()
         algo = self.get_field('a', decode=True)
         if algo.startswith('ed25519'):
             sdigest = DevsigHeader._validate_ed25519(bdata, keyinfo)
@@ -211,24 +222,24 @@ class DevsigHeader:
             signkey = keyinfo
             if not signtime:
                 raise ValidationError('t= field is required for ed25519 sigs')
+            if sdigest != vdigest:
+                raise ValidationError('Header validation failed')
+        elif algo.startswith('openssh'):
+            DevsigHeader._validate_openssh(bdata, vdigest, keyinfo)
+            signtime = self.get_field('t', decode=True)
+            signkey = keyinfo
+            if not signtime:
+                raise ValidationError('t= field is required for openssh sigs')
         elif algo.startswith('openpgp'):
             sdigest, (good, valid, trusted, signkey, signtime) = DevsigHeader._validate_openpgp(bdata, keyinfo)
+            if sdigest != vdigest:
+                raise ValidationError('Header validation failed')
         else:
             raise ValidationError('Unknown algorithm: %s', algo)
 
-        # Now we calculate our own digest
-        hashed = hashlib.sha256()
-        # Add in our _headervals first (they aready have CRLF endings)
-        hashed.update(b''.join(self._headervals))
-        # and the devsig header now, without the trailing CRLF
-        hashed.update(DEVSIG_HDR.lower() + b':' + dshdr)
-        vdigest = hashed.digest()
-        if sdigest != vdigest:
-            raise ValidationError('Header validation failed')
-
         return signkey, signtime
 
-    def sign(self, keyinfo: bytes, split: bool = True) -> Tuple[bytes, bytes]:
+    def sign(self, keyinfo: Union[str, bytes], split: bool = True) -> Tuple[bytes, bytes]:
         self.sanity_check()
         self.set_field('bh', self._body_hash)
         algo = self.get_field('a', decode=True)
@@ -251,6 +262,8 @@ class DevsigHeader:
             bval, pkinfo = DevsigHeader._sign_ed25519(digest, keyinfo)
         elif algo.startswith('openpgp'):
             bval, pkinfo = DevsigHeader._sign_openpgp(digest, keyinfo)
+        elif algo.startswith('openssh'):
+            bval, pkinfo = DevsigHeader._sign_openssh(digest, keyinfo)
         else:
             raise RuntimeError('Unknown a=%s' % algo)
 
@@ -295,7 +308,57 @@ class DevsigHeader:
             raise ValidationError('Failed to validate signature')
 
     @staticmethod
-    def _sign_openpgp(payload: bytes, keyid: bytes) -> Tuple[bytes, bytes]:
+    def _sign_openssh(payload: bytes, keyfile: str) -> Tuple[bytes, bytes]:
+        global KEYCACHE
+        keypath = os.path.expanduser(os.path.expandvars(keyfile))
+        if not os.access(keypath, os.R_OK):
+            raise SigningError('Unable to read openssh public key %s' % keypath)
+        sshkargs = ['-Y', 'sign', '-n', 'patatt', '-f', keypath]
+        ecode, out, err = sshk_run_command(sshkargs, payload)
+        if ecode > 0:
+            raise SigningError('Running ssh-keygen failed', errors=err.decode().split('\n'))
+        # Remove the header/footer
+        sigdata = b''
+        for bline in out.split(b'\n'):
+            if bline.startswith(b'----'):
+                continue
+            sigdata += bline
+        if keypath not in KEYCACHE:
+            # Now get the fingerprint of this keyid
+            sshkargs = ['-l', '-f', keypath]
+            ecode, out, err = sshk_run_command(sshkargs, payload)
+            if ecode > 0:
+                raise SigningError('Running ssh-keygen failed', errors=err.decode().split('\n'))
+            chunks = out.split()
+            keyfp = chunks[1]
+            KEYCACHE[keypath] = keyfp
+        else:
+            keyfp = KEYCACHE[keypath]
+
+        return sigdata, keyfp
+
+    @staticmethod
+    def _validate_openssh(sigdata: bytes, payload: bytes, keydata: bytes) -> None:
+        with tempfile.TemporaryDirectory(suffix='.patatt.ssh') as td:
+            # Start by making a signers file
+            fpath = os.path.join(td, 'signers')
+            spath = os.path.join(td, 'sigdata')
+            with open(fpath, 'wb') as fh:
+                chunks = keydata.split()
+                bcont = b'patatter@local namespaces="patatt" ' + chunks[0] + b' ' + chunks[1] + b'\n'
+                logger.debug('allowed-signers: %s', bcont)
+                fh.write(bcont)
+            with open(spath, 'wb') as fh:
+                bcont = b'-----BEGIN SSH SIGNATURE-----\n' + sigdata + b'\n-----END SSH SIGNATURE-----\n'
+                logger.debug('sigdata: %s', bcont)
+                fh.write(bcont)
+            sshkargs = ['-Y', 'verify', '-n', 'patatt', '-I', 'patatter@local', '-f', fpath, '-s', spath]
+            ecode, out, err = sshk_run_command(sshkargs, payload)
+            if ecode > 0:
+                raise ValidationError('Failed to validate openssh signature', errors=err.decode().split('\n'))
+
+    @staticmethod
+    def _sign_openpgp(payload: bytes, keyid: str) -> Tuple[bytes, bytes]:
         global KEYCACHE
         gpgargs = ['-s', '-u', keyid]
         ecode, out, err = gpg_run_command(gpgargs, payload)
@@ -331,11 +394,12 @@ class DevsigHeader:
         bsigdata = base64.b64decode(sigdata)
         vrfyargs = ['--verify', '--output', '-', '--status-fd=2']
         if pubkey:
-            with tempfile.TemporaryFile(suffix='.patch-attest-poc') as temp_keyring:
-                keyringargs = ['--no-default-keyring', f'--keyring={temp_keyring.name}']
+            with tempfile.TemporaryDirectory(suffix='.patatt.gnupg') as td:
+                keyringargs = ['--homedir', td, '--no-default-keyring', '--keyring', 'pub']
                 if pubkey in KEYCACHE:
                     logger.debug('Reusing cached keyring')
-                    temp_keyring.write(KEYCACHE[pubkey])
+                    with open(os.path.join(td, 'pub'), 'wb') as kfh:
+                        kfh.write(KEYCACHE[pubkey])
                 else:
                     logger.debug('Importing into new keyring')
                     gpgargs = keyringargs + ['--status-fd=1', '--import']
@@ -343,7 +407,8 @@ class DevsigHeader:
                     # look for IMPORT_OK
                     if out.find(b'[GNUPG:] IMPORT_OK') < 0:
                         raise ValidationError('Could not import GnuPG public key')
-                    KEYCACHE[pubkey] = temp_keyring.read()
+                    with open(os.path.join(td, 'pub'), 'rb') as kfh:
+                        KEYCACHE[pubkey] = kfh.read()
                 gpgargs = keyringargs + vrfyargs
                 ecode, out, err = gpg_run_command(gpgargs, stdin=bsigdata)
 
@@ -370,6 +435,7 @@ class DevsigHeader:
         signtime = ''
         signkey = ''
 
+        logger.debug('GNUPG status:\n\t%s', status.decode().strip().replace('\n', '\n\t'))
         gs_matches = re.search(rb'^\[GNUPG:] GOODSIG ([0-9A-F]+)\s+(.*)$', status, flags=re.M)
         if gs_matches:
             good = True
@@ -487,17 +553,18 @@ class PatattMessage:
         ds.set_headers(self.canon_headers, mode='sign')
         ds.set_body(self.canon_body)
         ds.set_field('l', str(len(self.canon_body)))
-        if identity and identity != self.canon_identity:
-            ds.set_field('i', identity)
+        if not identity:
+            identity = self.canon_identity
+        ds.set_field('i', identity)
         if selector:
             ds.set_field('s', selector)
 
-        if algo not in ('ed25519', 'openpgp'):
+        if algo not in ('ed25519', 'openpgp', 'openssh'):
             raise SigningError('Unsupported algorithm: %s' % algo)
 
         ds.set_field('a', '%s-sha256' % algo)
-        if algo == 'ed25519':
-            # Set signing time for ed25519 sigs
+        if algo in ('ed25519', 'openssh'):
+            # Set signing time for non-pgp sigs
             ds.set_field('t', str(int(time.time())))
         hv, pkinfo = ds.sign(keyinfo)
 
@@ -511,6 +578,8 @@ class PatattMessage:
         ]
         if algo == 'openpgp':
             idata.append(b'fpr=%s' % pkinfo)
+        elif algo == 'openssh':
+            idata.append(b'fpr=%s' % pkinfo)
         else:
             idata.append(b'pk=%s' % pkinfo)
 
@@ -611,7 +680,9 @@ class PatattMessage:
             mf = os.path.join(td, 'm')
             pf = os.path.join(td, 'p')
             cmdargs = ['git', 'mailinfo', '--encoding=utf-8', '--no-scissors', mf, pf]
-            ecode, i, err = _run_command(cmdargs, stdin=payload)
+            # normalize line endings in payload to single lf for consistent results
+            # from git-mailinfo.
+            ecode, i, err = _run_command(cmdargs, stdin=payload.replace(b'\r\n', b'\n'))
             if ecode > 0:
                 logger.debug('FAILED  : Failed running git-mailinfo:')
                 logger.debug(err.decode())
@@ -701,10 +772,17 @@ def get_config_from_git(regexp: str, section: Optional[str] = None, defaults: Op
 
 
 def gpg_run_command(cmdargs: list, stdin: bytes = None) -> Tuple[int, bytes, bytes]:
+    set_bin_paths(None)
     cmdargs = [GPGBIN, '--batch', '--no-auto-key-retrieve', '--no-auto-check-trustdb'] + cmdargs
     return _run_command(cmdargs, stdin)
 
 
+def sshk_run_command(cmdargs: list, stdin: bytes = None) -> Tuple[int, bytes, bytes]:
+    set_bin_paths(None)
+    cmdargs = [SSHKBIN] + cmdargs
+    return _run_command(cmdargs, stdin)
+
+
 def get_git_toplevel(gitdir: str = None) -> str:
     cmdargs = ['git']
     if gitdir:
@@ -833,12 +911,37 @@ def sign_message(msgdata: bytes, algo: str, keyinfo: Union[str, bytes],
     return pm.as_bytes()
 
 
-def cmd_sign(cmdargs, config: dict) -> None:
+def set_bin_paths(config: Optional[dict]) -> None:
+    global GPGBIN, SSHKBIN
+    if GPGBIN is None:
+        gpgcfg = get_config_from_git(r'gpg\..*')
+        if config and config.get('gpg-bin'):
+            GPGBIN = config.get('gpg-bin')
+        elif gpgcfg.get('program'):
+            GPGBIN = gpgcfg.get('program')
+        else:
+            GPGBIN = 'gpg'
+    if SSHKBIN is None:
+        sshcfg = get_config_from_git(r'gpg\..*', section='ssh')
+        if config and config.get('ssh-keygen-bin'):
+            SSHKBIN = config.get('ssh-keygen-bin')
+        elif sshcfg.get('program'):
+            SSHKBIN = sshcfg.get('program')
+        else:
+            SSHKBIN = 'ssh-keygen'
+
+
+def get_algo_keydata(config: dict) -> Tuple[str, str]:
+    global KEYCACHE
     # Do we have the signingkey defined?
     usercfg = get_config_from_git(r'user\..*')
     if not config.get('identity') and usercfg.get('email'):
         # Use user.email
         config['identity'] = usercfg.get('email')
+    # Do we have this already looked up?
+    if config['identity'] in KEYCACHE:
+        return KEYCACHE[config['identity']]
+
     if not config.get('signingkey'):
         if usercfg.get('signingkey'):
             logger.info('N: Using pgp key %s defined by user.signingkey', usercfg.get('signingkey'))
@@ -847,13 +950,7 @@ def cmd_sign(cmdargs, config: dict) -> None:
         else:
             logger.critical('E: patatt.signingkey is not set')
             logger.critical('E: Perhaps you need to run genkey first?')
-            sys.exit(1)
-
-    try:
-        messages = _load_messages(cmdargs)
-    except IOError as ex:
-        logger.critical('E: %s', ex)
-        sys.exit(1)
+            raise NoKeyError('patatt.signingkey is not set')
 
     sk = config.get('signingkey')
     if sk.startswith('ed25519:'):
@@ -877,8 +974,7 @@ def cmd_sign(cmdargs, config: dict) -> None:
                         keysrc = skey
 
         if not keysrc:
-            logger.critical('E: Could not find the key matching %s', identifier)
-            sys.exit(1)
+            raise ConfigurationError('Could not find the key matching %s' % identifier)
 
         logger.info('N: Using ed25519 key: %s', keysrc)
         with open(keysrc, 'r') as fh:
@@ -887,24 +983,67 @@ def cmd_sign(cmdargs, config: dict) -> None:
     elif sk.startswith('openpgp:'):
         algo = 'openpgp'
         keydata = sk[8:]
+    elif sk.startswith('openssh:'):
+        algo = 'openssh'
+        keydata = sk[8:]
     else:
         logger.critical('E: Unknown key type: %s', sk)
+        raise ConfigurationError('Unknown key type: %s' % sk)
+
+    KEYCACHE[config['identity']] = (algo, keydata)
+    return algo, keydata
+
+
+def rfc2822_sign(message: bytes, config: Optional[dict] = None) -> bytes:
+    if config is None:
+        config = get_main_config()
+    algo, keydata = get_algo_keydata(config)
+    pm = PatattMessage(message)
+    pm.sign(algo, keydata, identity=config.get('identity'), selector=config.get('selector'))
+    logger.debug('--- SIGNED MESSAGE STARTS ---')
+    logger.debug(pm.as_string())
+    return pm.as_bytes()
+
+
+def get_main_config(section: Optional[str] = None) -> dict:
+    global CONFIGCACHE
+    if section in CONFIGCACHE:
+        return CONFIGCACHE[section]
+    config = get_config_from_git(r'patatt\..*', section=section, multivals=['keyringsrc'])
+    # Append some extra keyring locations
+    if 'keyringsrc' not in config:
+        config['keyringsrc'] = list()
+    config['keyringsrc'] += ['ref:::.keys', 'ref:::.local-keys', 'ref::refs/meta/keyring:']
+    set_bin_paths(config)
+    logger.debug('config: %s', config)
+    CONFIGCACHE[section] = config
+    return config
+
+
+def cmd_sign(cmdargs, config: dict) -> None:
+    try:
+        messages = _load_messages(cmdargs)
+    except IOError as ex:
+        logger.critical('E: %s', ex)
         sys.exit(1)
 
     for fn, msgdata in messages.items():
         try:
-            pm = PatattMessage(msgdata)
-            pm.sign(algo, keydata, identity=config.get('identity'), selector=config.get('selector'))
+            signed = rfc2822_sign(msgdata, config)
             logger.debug('--- SIGNED MESSAGE STARTS ---')
-            logger.debug(pm.as_string())
+            logger.debug(signed.decode())
             if fn == '-':
-                sys.stdout.buffer.write(pm.as_bytes())
+                sys.stdout.buffer.write(signed)
             else:
                 with open(fn, 'wb') as fh:
-                    fh.write(pm.as_bytes())
+                    fh.write(signed)
 
                 logger.critical('SIGN | %s', os.path.basename(fn))
 
+        except (ConfigurationError, NoKeyError) as ex:
+            logger.debug('Exiting due to %s', ex)
+            sys.exit(1)
+
         except SigningError as ex:
             logger.critical('E: %s', ex)
             sys.exit(1)
@@ -919,6 +1058,7 @@ def validate_message(msgdata: bytes, sources: list, trim_body: bool = False) ->
     pm = PatattMessage(msgdata)
     if not pm.signed:
         logger.debug('message is not signed')
+        attestations.append((RES_NOSIG, None, None, None, None, ['no signatures found']))
         return attestations
 
     # Find all identities for which we have public keys
@@ -934,6 +1074,8 @@ def validate_message(msgdata: bytes, sources: list, trim_body: bool = False) ->
             algo = 'ed25519'
         elif a.startswith('openpgp'):
             algo = 'openpgp'
+        elif a.startswith('openssh'):
+            algo = 'openssh'
         else:
             errors.append('%s/%s Unknown algorigthm: %s' % (i, s, a))
             attestations.append((RES_ERROR, i, t, None, a, errors))
@@ -947,8 +1089,8 @@ def validate_message(msgdata: bytes, sources: list, trim_body: bool = False) ->
             except KeyError:
                 pass
 
-        if not pkey and algo == 'ed25519':
-            errors.append('%s/%s no matching ed25519 key found' % (i, s))
+        if not pkey and algo in ('ed25519', 'openssh'):
+            errors.append('%s/%s no matching %s key found' % (i, s, algo))
             attestations.append((RES_NOKEY, i, t, None, algo, errors))
             continue
 
@@ -1019,6 +1161,10 @@ def cmd_validate(cmdargs, config: dict):
                         logger.info('       | key: %s', keysrc)
                     else:
                         logger.info('       | key: default GnuPG keyring')
+                elif result <= RES_NOSIG:
+                    logger.critical(' NOSIG | %s', fn)
+                    for error in errors:
+                        logger.critical('       | %s', error)
                 elif result <= RES_NOKEY:
                     logger.critical(' NOKEY | %s, %s', identity, fn)
                     for error in errors:
@@ -1190,12 +1336,7 @@ def command() -> None:
         ch.setLevel(logging.CRITICAL)
 
     logger.addHandler(ch)
-    config = get_config_from_git(r'patatt\..*', section=_args.section, multivals=['keyringsrc'])
-    # Append some extra keyring locations
-    if 'keyringsrc' not in config:
-        config['keyringsrc'] = list()
-    config['keyringsrc'] += ['ref:::.keys', 'ref:::.local-keys', 'ref::refs/meta/keyring:']
-    logger.debug('config: %s', config)
+    config = get_main_config(section=_args.section)
 
     if 'func' not in _args:
         parser.print_help()
